<?php

namespace Hej\CTI;

use BeanFactory;
use SugarBean;
use SugarEmailAddress;

class HejSalesCti
{
    protected static $FETCH_FIELDS = [
        "id",
        "title",
        "first_name",
        "last_name",
        "phone_home",
        "phone_work",
        "phone_mobile",
        "phone_other",
        "phone_fax",
        "account_name",
    ];

    /**
     * Search in contacts and leads by name
     *
     * @param string $name
     * @return array
     */
    public static function findContactByName($name)
    {
        $data = [];

        if ($name) {
            $contact = BeanFactory::newBean("Contacts");
            $contacts = static::executeNameSearchQuery($contact, $name);

            $lead = BeanFactory::newBean("Leads");
            $leads = static::executeNameSearchQuery($lead, $name);

            $data = array_merge($contacts, $leads);
        }

        return $data;
    }

    /**
     * Search in contacts and leads by phone name
     *
     * @param string $phone
     * @return array
     */
    public static function findContactByPhone($phone)
    {
        $data = [];

        if (static::validatePhoneQuery($phone)) {
            $countryCode = substr($phone, 1, 2);
            $number = str_split(substr($phone, 3));

            $contact = BeanFactory::newBean("Contacts");
            $contacts = static::executePhoneSearchQuery($contact, $countryCode, $number);

            $lead = BeanFactory::newBean("Leads");
            $leads = static::executePhoneSearchQuery($lead, $countryCode, $number);

            $data = array_merge($contacts, $leads);
        }

        return $data;
    }

    /**
     * Search in contacts and leads by email address
     *
     * @param string $email
     * @return array
     */
    public static function findContactByEmail($email)
    {
        $data = [];

        if (static::validateEmailQuery($email)) {
            $contact = BeanFactory::newBean("Contacts");
            $contacts = static::executeEmailSearchQuery($contact, $email);

            $lead = BeanFactory::newBean("Leads");
            $leads = static::executeEmailSearchQuery($lead, $email);

            $data = array_merge($contacts, $leads);
        }

        return $data;
    }

    /**
     * Execute name search query for the given bean
     *
     * @param SugarBean $bean
     * @param string $query
     * @return array
     */
    protected static function executeNameSearchQuery(SugarBean $bean, $query)
    {
        global $db;
        $query = $db->quoted("%$query%");

        $where = "CONCAT({$bean->table_name}.first_name, ' ', {$bean->table_name}.last_name) LIKE $query OR " .
            "CONCAT({$bean->table_name}.last_name, ' ', {$bean->table_name}.first_name) LIKE $query";

        $fields = static::$FETCH_FIELDS;
        static::beanSpecificQueryOpts($bean, $where, $fields);

        $beans = $bean->get_list(
            null,
            $where,
            0,
            20,
            -1,
            false,
            $fields
        );

        return static::extractBeanList($beans["list"], $fields);
    }

    /**
     * Execute a phone search query for the given bean
     *
     * @param SugarBean $bean
     * @param $countryCode
     * @param array $number
     * @return array
     */
    protected static function executePhoneSearchQuery(SugarBean $bean, $countryCode, array $number)
    {
        $phoneRegex = static::createPhoneQueryRegex($countryCode, $number);
        $where = "";

        foreach (static::$FETCH_FIELDS as $field) {
            if (strpos($field, "phone_") !== false) {
                if (strlen($where) > 0) {
                    $where .= " OR ";
                }

                $where .= "{$bean->table_name}.$field REGEXP '$phoneRegex'";
            }
        }

        $fields = static::$FETCH_FIELDS;
        static::beanSpecificQueryOpts($bean, $where, $fields);

        $beans = $bean->get_list(
            null,
            $where,
            0,
            20,
            -1,
            false,
            $fields
        );

        return static::extractBeanList($beans["list"], $fields);
    }

    /**
     * Execute an email search query for the given bean
     *
     * @param SugarBean $bean
     * @param string $email
     * @return array
     */
    protected static function executeEmailSearchQuery(SugarBean $bean, $email)
    {
        $sea = new SugarEmailAddress();
        $beans = $sea->getBeansByEmailAddress($email);

        $relevant = array_filter($beans, function ($b) use ($bean) {
            return $b->module_name === $bean->module_name;
        });

        $ids = array_map(function ($b) {
            global $db;

            return $db->quoted($b->id);
        }, $relevant);

        if (count($ids) === 0) {
            return [];
        }

        $idWhere = implode(",", $ids);
        $where = "`{$bean->table_name}`.`id` IN ({$idWhere})";

        $fields = static::$FETCH_FIELDS;
        static::beanSpecificQueryOpts($bean, $where, $fields);

        $beans = $bean->get_list(
            null,
            $where,
            0,
            20,
            -1,
            false,
            $fields
        );

        return static::extractBeanList($beans["list"], $fields);
    }

    /**
     * Extends the where clause and fields list with bean specific options
     *
     * @param SugarBean $bean
     * @param string $where
     * @param array $fields
     */
    protected static function beanSpecificQueryOpts(SugarBean $bean, &$where, array &$fields)
    {
        if ($bean->module_name === "Contacts") {
            $fields[] = "account_id";
        } else if ($bean->module_name === "Leads") {
            $where = "{$bean->table_name}.converted = 0 AND ($where)";
        }
    }

    /**
     * Extract the beans from an array of SugarBeans into the required response DTOs
     *
     * @param SugarBean[] $beans
     * @param string[] $fields
     * @return array
     */
    protected static function extractBeanList(array $beans, array $fields)
    {
        $idDedupe = [];
        $result = [];
        foreach ($beans as $bean) {
            if (in_array($bean->id, $idDedupe)) {
                /*
                 * Due to a bug in Suite the same bean can be returned multiple times.
                 * We just deduplicate them here
                 */
                continue;
            }

            $idDedupe[] = $bean->id;

            $arr = array_filter($bean->toArray(), function ($key) use ($fields) {
                return in_array($key, $fields);
            }, ARRAY_FILTER_USE_KEY);

            $arr["id"] = "{$bean->module_name}:{$bean->id}";

            if (!$arr["account_id"]) {
                $arr["account_id"] = null;
            }

            $arr["email"] = static::getEmailAddressesFor($bean);

            $result[] = $arr;
        }

        return $result;
    }

    /**
     * Validate the given query as a E.164 phone number
     *
     * @param string $query
     * @return bool
     */
    protected static function validatePhoneQuery($query)
    {
        return preg_match('/^\\+[0-9]+$/', $query) === 1;
    }

    /**
     * Validate the given query to be an email address
     *
     * @param string $query
     * @return bool
     */
    protected static function validateEmailQuery($query)
    {
        return preg_match(
                '/^(?!(?:(?:\x22?\x5C[\x00-\x7E]\x22?)|(?:\x22?[^\x5C\x22]\x22?)){255,})(?!(?:(?:\x22?\x5C[\x00-\x7E]\x22?)|(?:\x22?[^\x5C\x22]\x22?)){65,}@)(?:(?:[\x21\x23-\x27\x2A\x2B\x2D\x2F-\x39\x3D\x3F\x5E-\x7E]+)|(?:\x22(?:[\x01-\x08\x0B\x0C\x0E-\x1F\x21\x23-\x5B\x5D-\x7F]|(?:\x5C[\x00-\x7F]))*\x22))(?:\.(?:(?:[\x21\x23-\x27\x2A\x2B\x2D\x2F-\x39\x3D\x3F\x5E-\x7E]+)|(?:\x22(?:[\x01-\x08\x0B\x0C\x0E-\x1F\x21\x23-\x5B\x5D-\x7F]|(?:\x5C[\x00-\x7F]))*\x22)))*@(?:(?:(?!.*[^.]{64,})(?:(?:(?:xn--)?[a-z0-9]+(?:-[a-z0-9]+)*\.){1,126}){1,}(?:(?:[a-z][a-z0-9]*)|(?:(?:xn--)[a-z0-9]+))(?:-[a-z0-9]+)*)|(?:\[(?:(?:IPv6:(?:(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){7})|(?:(?!(?:.*[a-f0-9][:\]]){7,})(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){0,5})?::(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){0,5})?)))|(?:(?:IPv6:(?:(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){5}:)|(?:(?!(?:.*[a-f0-9]:){5,})(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){0,3})?::(?:[a-f0-9]{1,4}(?::[a-f0-9]{1,4}){0,3}:)?)))?(?:(?:25[0-5])|(?:2[0-4][0-9])|(?:1[0-9]{2})|(?:[1-9]?[0-9]))(?:\.(?:(?:25[0-5])|(?:2[0-4][0-9])|(?:1[0-9]{2})|(?:[1-9]?[0-9]))){3}))\]))$/iD',
                $query
            ) === 1;
    }

    /**
     * Create a MySQL regular expression for phone number search
     *
     * @param string $countryCode
     * @param array $number
     * @return string
     */
    protected static function createPhoneQueryRegex($countryCode, array $number)
    {
        return '^(\\\+|00)?(' . $countryCode . ')?(\\\(?0\\\)?)?.*' . join(".*", $number) . "$";
    }

    /**
     * Get list of email addresses for given bean
     *
     * @param SugarBean $bean
     * @return array
     */
    protected static function getEmailAddressesFor(SugarBean $bean)
    {
        $sea = new SugarEmailAddress();
        $fetched = $sea->getAddressesByGUID($bean->id, $bean->module_name);

        return array_values(array_map(function ($email) {
            return $email["email_address"];
        }, $fetched));
    }
}
